module Logic where

import Maybe 
import Data.Generics hiding ((:*:),Prefix)

import StrategyLib
import TermRep

import ApplyStrategy
import ParserCombinatorsWithStateBFLib

--------------------------------------------------------------------------------
-- Logic domain
--------------------------------------------------------------------------------

data Logic = LVar String
           | Logic :->:  Logic            -- implication
           | Logic :<->: Logic            -- equivalence
           | Logic :&&:  Logic            -- and (conjunction)
           | Logic :||:  Logic            -- or (disjunction)
           | Not Logic                    -- not
           | T                            -- true
           | F                            -- false    
             deriving (Show,Eq,Ord,Data,Typeable)

--------------------------------------------------------------------------------
-- Zipper for Logic
--------------------------------------------------------------------------------

data Ctx = Top
         | ImplL Ctx Logic
         | ImplR Logic Ctx
         | EquivL Ctx Logic
         | EquivR Logic Ctx
         | AndL Ctx Logic
         | AndR Logic Ctx
         | OrL Ctx Logic
         | OrR Logic Ctx
         | NotD Ctx 
           deriving Show

data Direction = DImplL
               | DImplR
               | DEquivL
               | DEquivR
               | DAndL
               | DAndR
               | DOrL
               | DOrR
               | DNotD
               | DUp
               | DEmpty
   deriving (Show,Eq,Ord,Data,Typeable)

move :: Direction -> Loc Logic Ctx -> Maybe (Loc Logic Ctx)
move dir loc = case dir of
                 DImplL     ->  case it loc of
                                  a :->: b   ->  Just (At a (ImplL (ctx loc) b))
                                  otherwise  ->  Nothing 
                 DImplR     ->  case it loc of
                                  a :->: b   ->  Just (At b (ImplR a (ctx loc)))
                                  otherwise  ->  Nothing 
                 DEquivL    ->  case it loc of
                                  a :<->: b  ->  Just (At a (EquivL (ctx loc) b))
                                  otherwise  ->  Nothing 
                 DEquivR    ->  case it loc of
                                  a :<->: b  ->  Just (At b (EquivR a (ctx loc)))
                                  otherwise  ->  Nothing 
                 DAndL      ->  case it loc of
                                  a :&&: b   ->  Just (At a (AndL (ctx loc) b))
                                  otherwise  ->  Nothing 
                 DAndR      ->  case it loc of
                                  a :&&: b   ->  Just (At b (AndR a (ctx loc)))
                                  otherwise  ->  Nothing 
                 DOrL       ->  case it loc of
                                  a :||: b   ->  Just (At a (OrL (ctx loc) b))
                                  otherwise  ->  Nothing 
                 DOrR       ->  case it loc of
                                  a :||: b   ->  Just (At b (OrR a (ctx loc)))
                                  otherwise  ->  Nothing 
                 DNotD      ->  case it loc of
                                  Not a      ->  Just (At a (NotD (ctx loc)))
                                  otherwise  ->  Nothing 
                 DUp        ->  case ctx loc of
                                  ImplL ctx r   ->  Just (At (it loc :->: r) ctx)
                                  ImplR l ctx   ->  Just (At (l :->: it loc) ctx)
                                  EquivL ctx r  ->  Just (At (it loc :<->: r) ctx)
                                  EquivR l ctx  ->  Just (At (l :<->: it loc) ctx)
                                  AndL ctx r    ->  Just (At (it loc :&&: r) ctx)
                                  AndR l ctx    ->  Just (At (l :&&: it loc) ctx)
                                  OrL ctx r     ->  Just (At (it loc :||: r) ctx)
                                  OrR l ctx     ->  Just (At (l :||: it loc) ctx)
                                  NotD ctx      ->  Just (At (it loc) ctx)
                                  otherwise     ->  Just loc
                 otherwise  ->  error (show dir ++ show loc)
                 
--------------------------------------------------------------------------------
-- Rules for Logic
--------------------------------------------------------------------------------

data TLogic  =  FalseUnitOr
             |  TrueZeroOr
             |  TrueUnitAnd
             |  FalseZeroAnd
             |  DeMorganOr 
             |  DeMorganAnd 
             |  NotBoolConst
             |  NotNot 
             |  AndOverOr
             |  LMove Direction
             |  TLogicQED 
                deriving (Show,Eq)

instance Symbol TLogic where
  symQed             =  TLogicQED
  isMove (LMove _)   =  True
  isMove _           =  False
  penalty TLogicQED  =  25
  penalty s          =  5

makeRule r  =  adhocTP idTP r


falseUnitOr = makeRule ruleFalseUnitOr

ruleFalseUnitOr (F :||: x)  =  return x
ruleFalseUnitOr (x :||: F)  =  return x
ruleFalseUnitOr x           =  fail "FalseUnitOr"
  
testFUO1 :: Maybe Logic
testFUO1 = applyS falseUnitOr (F :||: LVar "OK")

testFUO2 :: Maybe Logic
testFUO2 = applyS falseUnitOr (F :&&: LVar "OK")


trueZeroOr = makeRule ruleTrueZeroOr

ruleTrueZeroOr (T :||: x)  =  return T
ruleTrueZeroOr (x :||: T)  =  return T
ruleTrueZeroOr x           =  fail "TrueZeroOr"


trueUnitAnd = makeRule ruleTrueUnitAnd

ruleTrueUnitAnd (T :&&: x)  =  return x
ruleTrueUnitAnd (x :&&: T)  =  return x
ruleTrueUnitAnd x           =  fail "TrueUnitAnd"


falseZeroAnd = makeRule ruleFalseZeroAnd

ruleFalseZeroAnd (F :&&: x)  =  return F
ruleFalseZeroAnd (x :&&: F)  =  return F
ruleFalseZeroAnd x           =  fail "FalseZeroAnd"


deMorganOr = makeRule ruleDeMorganOr

ruleDeMorganOr (Not (x :||: y)) = return (Not x :&&: Not y)
ruleDeMorganOr x                = fail "DeMorganOr"


deMorganAnd = makeRule ruleDeMorganAnd

ruleDeMorganAnd (Not (x :&&: y)) = return (Not x :||: Not y)
ruleDeMorganAnd x                = fail "DeMorganAnd"


notBoolConst = makeRule ruleNotBoolConst

ruleNotBoolConst (Not T) = return F
ruleNotBoolConst (Not F) = return T
ruleNotBoolConst x       = fail "NotBoolConst"


notNot = makeRule ruleNotNot

ruleNotNot (Not (Not x)) = return x
ruleNotNot x             = fail "NotNot"


andOverOr = makeRule ruleAndOverOr

ruleAndOverOr (x :&&: (y :||: z)) = return ((x :&&: y) :||: (x :&&: z))
ruleAndOverOr ((x :||: y) :&&: z) = return ((x :&&: y) :||: (y :&&: z))
ruleAndOverOr x                   = fail "AndOverOr"


tryApply rule loc = case applyS rule (it loc) of 
                      Just domain' -> Just (At domain' (ctx loc))
                      Nothing      -> Nothing

instance Apply TLogic (Loc Logic Ctx) where
  apply step  = 
    case step of
      FalseUnitOr  -> tryApply falseUnitOr
      TrueZeroOr   -> tryApply trueZeroOr
      TrueUnitAnd  -> tryApply trueUnitAnd
      FalseZeroAnd -> tryApply falseZeroAnd
      DeMorganAnd  -> tryApply deMorganAnd 
      DeMorganOr   -> tryApply deMorganOr 
      NotBoolConst -> tryApply notBoolConst
      NotNot       -> tryApply notNot 
      AndOverOr    -> tryApply andOverOr
      LMove dir    -> move dir
      TLogicQED    -> Just

instance Children Logic where
  children (a :&&: b) = [a,b]
  children (a :||: b) = [a,b]
  children (Not a) = [a]
  children (LVar _) = []
